#include <STDLIB.H>
#include <STDIO.H>
#include <DOS.H>

void setPixel(int AX, int AY, unsigned char AC) {
  pokeb(0xA000, AY*320+AX, AC);
}

void setScreenMode(void) {
  int i;
  static const unsigned char pal[] =
  {  0,   0,   0,
     0, 168, 168,
   252,  84, 252,
   252, 252, 252,
     0,   0, 168,
    84, 252,  84,
   252,  84,  84,
   252, 252,  84,
    84,  84,  84,
    84,  84, 252,
   168,   0, 168,
    84, 252, 252,
   168,  84,   0,
     0, 168,   0,
   168,   0,   0,
   168, 168, 168
  };

  asm {
    mov ax, 0x0013;
    int 0x10;
  }

  //set palette
  outportb(0x3C8, 0);
  for (i = 0; i < 16*3; i++) {
    outportb(0x3C9, pal[i] >> 2);
  }
}

int main(int argc, char* argv[]) {
  unsigned char mask;
  int i, j;
  int x, y;
  int pel;
  FILE* spriteFile;
  char* spriteData;
  typedef struct {
    char FUnknown[12];
    unsigned char FWidth;
    unsigned char FHeight;
  } headerType;
  headerType header;

  if (argc < 2) {
    printf("Mega Man III sprite viewer by Frenkel, 23-Aug-2008.\n");
    printf("Usage: tiles DECOMPRESSEDSPRITESFILE\n");
    return -1;
  }

  spriteFile = fopen(argv[1], "rb");
  if (spriteFile == NULL) {
    printf("Mega Man III sprite viewer by Frenkel, 23-Aug-2008.\n");
    printf("Can't find %s!\n", argv[1]);
    return -1;
  }

  setScreenMode();

  //Draw sprite
  i = fread(&header, sizeof(headerType), 1, spriteFile);
  while (i != 0) {
    //clear screen
    for (y = 0; y < 200; y++) {
      for (x = 0; x < 320; x++) {
	setPixel(x, y, 67);
      }
    }

    //allocate memory
    spriteData = (char *) malloc(header.FHeight * header.FWidth * (1+8));
    if (spriteData == NULL) {
      asm {
	mov ax, 0x0003;
	int 0x10;
      }
      printf("Mega Man III sprite viewer by Frenkel, 23-Aug-2008.\n");
      printf("Not enough memory available!\n");
      return -1;
    }

    //get sprite data from sprite file
    fread(spriteData, header.FHeight * header.FWidth * (1+8), 1, spriteFile);

    //draw sprite
    j = 0;
    for (y = 0; y < header.FHeight; y++) {
      for (pel = 0; pel < header.FWidth; pel++) {
	mask = spriteData[j];
	j++;
	for (x = 0; x < 8; x++) {
	  if (mask & (128 >> x)) {
	    setPixel(x + pel * 8, y, spriteData[j]);
	  }
	  j++;
	}
      }
    }
    free(spriteData);

    asm {
      xor ah, ah;
      int 0x16;
    }
    i = fread(&header, sizeof(headerType), 1, spriteFile);
  }

  fclose(spriteFile);

  asm {
    mov ax, 0x0003;
    int 0x10;
  }

  printf("Mega Man III sprite viewer by Frenkel, 23-Aug-2008.\n");
  return 0;
}